@@ -17,7 +17,7 @@ from mangoapi import get_chapter, get_title, search_title
 
 from . import mangadex
 from .conf import config
-from .decorators import require_login
+from .decorators import require_login, trigger_has_read
 from .persistence import (
     follow,
     get_followed_titles,
@@ -155,6 +155,7 @@ def auth_view():
 
 
 @app.route("/title/<site>/<title_id>")
+@trigger_has_read
 def title_view(site, title_id):
     user = session.get("user", None)
     user_id = user["id"] if user else None
@@ -171,6 +172,7 @@ def title_view(site, title_id):
 
 
 @app.route("/chapter/<site>/<chapter_id>")
+@trigger_has_read
 def chapter_view(site, chapter_id):
     chapter = load_chapter(site, chapter_id)
     if not chapter:
