@@ -3,12 +3,11 @@ import json
 import apsw
 import argon2
 
-from .database.common import get_conn, run_sql
+from .database.common import run_sql, run_sql_on_demand
 
 
 def save_title(title):
-    conn = get_conn()
-    conn.cursor().execute(
+    run_sql(
         """
     INSERT INTO title (
         id,
@@ -48,18 +47,15 @@ def save_title(title):
 
 
 def load_title(site, title_id, user_id=None):
-    conn = get_conn()
-    result = list(
-        conn.cursor().execute(
-            """
-    SELECT id, name, site, cover_ext, chapters, alt_names, descriptions
-    FROM title
-    WHERE id = ?
-      AND site = ?
-      AND datetime(updated_at) > datetime('now', '-6 hours');
-    """,
-            (title_id, site),
-        )
+    result = run_sql(
+        """
+        SELECT id, name, site, cover_ext, chapters, alt_names, descriptions
+        FROM title
+        WHERE id = ?
+          AND site = ?
+          AND datetime(updated_at) > datetime('now', '-6 hours');
+        """,
+        (title_id, site),
     )
     if not result:
         return None
@@ -68,29 +64,38 @@ def load_title(site, title_id, user_id=None):
     else:
         title = result[0]
 
-        return_val = {
-            "id": title[0],
-            "name": title[1],
-            "site": title[2],
-            "cover_ext": title[3],
-            "chapters": json.loads(title[4]),
-            "alt_names": json.loads(title[5]),
-            "descriptions": json.loads(title[6]),
-        }
+        for field in ["chapters", "alt_names", "descriptions"]:
+            title[field] = json.loads(title[field])
 
         if user_id is not None:
-            return_val["is_following"] = bool(
+            title["is_following"] = bool(
                 run_sql(
                     "SELECT 1 FROM follow WHERE user_id=? AND site=? AND title_id=?;",
-                    (user_id, site, return_val["id"]),
+                    (user_id, site, title["id"]),
                 )
             )
-        return return_val
+
+            chapters_i_read = run_sql(
+                """
+                SELECT r.chapter_id
+                FROM read r
+                  INNER JOIN chapter c ON c.id = r.chapter_id AND c.site = r.site
+                WHERE r.user_id = ?
+                  AND c.title_id = ?
+                  AND c.site = ?
+                ORDER BY r.updated_at;
+                """,
+                (user_id, title["id"], title["site"]),
+            )
+
+            for ch in title["chapters"]:
+                if ch["id"] in chapters_i_read:
+                    ch["is_read"] = True
+        return title
 
 
 def save_chapter(chapter):
-    conn = get_conn()
-    conn.cursor().execute(
+    run_sql(
         """
     INSERT INTO chapter (
         id,
@@ -118,8 +123,8 @@ def save_chapter(chapter):
             "id": chapter["id"],
             "title_id": chapter["title_id"],
             "site": "mangadex",
-            "num_major": chapter["num_major"],
-            "num_minor": chapter.get("num_minor", None),
+            "num_major": chapter.get("num_major"),
+            "num_minor": chapter.get("num_minor"),
             "name": chapter["name"],
             "pages": json.dumps(chapter["pages"]),
             "groups": json.dumps(chapter["groups"]),
@@ -129,16 +134,13 @@ def save_chapter(chapter):
 
 
 def load_chapter(site, chapter_id):
-    conn = get_conn()
-    result = list(
-        conn.cursor().execute(
-            """
-    SELECT id, title_id, num_major, num_minor, name, pages, groups, is_webtoon
-    FROM chapter
-    WHERE id = ? AND site=?;
-    """,
-            (chapter_id, site),
-        )
+    result = run_sql(
+        """
+        SELECT id, title_id, num_major, num_minor, name, pages, groups, is_webtoon
+        FROM chapter
+        WHERE id = ? AND site=?;
+        """,
+        (chapter_id, site),
     )
     if not result:
         return None
@@ -146,16 +148,9 @@ def load_chapter(site, chapter_id):
         raise Exception(f"Found multiple results for chapter_id {chapter_id}!")
     else:
         chapter = result[0]
-        return {
-            "id": chapter[0],
-            "title_id": chapter[1],
-            "num_major": chapter[2],
-            "num_minor": chapter[3],
-            "name": chapter[4],
-            "pages": json.loads(chapter[5]),
-            "groups": json.loads(chapter[6]),
-            "is_webtoon": chapter[7],
-        }
+        chapter["pages"] = json.loads(chapter["pages"])
+        chapter["groups"] = json.loads(chapter["groups"])
+        return chapter
 
 
 def get_prev_next_chapters(title, chapter):
@@ -181,7 +176,7 @@ def register_user(username, password):
     hasher = argon2.PasswordHasher()
     hashed_password = hasher.hash(password)
     try:
-        get_conn().cursor().execute(
+        run_sql(
             "INSERT INTO user (username, password) VALUES (?, ?);",
             (username, hashed_password),
         )
@@ -193,18 +188,13 @@ def register_user(username, password):
 
 
 def verify_username_password(username, password):
-    data = list(
-        get_conn()
-        .cursor()
-        .execute("SELECT id, password FROM user WHERE username = ?;", (username,))
-    )
+    data = run_sql("SELECT id, password FROM user WHERE username = ?;", (username,))
     if len(data) != 1:
         print(f"User {username} doesn't exist.")
         return None
 
-    user_id = data[0][0]
-    hashed_password = data[0][1]
-
+    user_id = data[0]["id"]
+    hashed_password = data[0]["password"]
     hasher = argon2.PasswordHasher()
     try:
         hasher.verify(hashed_password, password)
@@ -215,14 +205,14 @@ def verify_username_password(username, password):
 
 
 def follow(user_id, site, title_id):
-    get_conn().cursor().execute(
+    run_sql(
         "INSERT INTO follow (user_id, site, title_id) VALUES (?, ?, ?);",
         (user_id, site, title_id),
     )
 
 
 def unfollow(user_id, site, title_id):
-    get_conn().cursor().execute(
+    run_sql(
         "DELETE FROM follow WHERE user_id=? AND site=? AND title_id=?;",
         (user_id, site, title_id),
     )
@@ -239,11 +229,51 @@ def get_followed_titles(user_id):
         """,
         (user_id,),
     )
-    keys = ("id", "site", "name", "cover_ext", "chapters")
-    title_dicts = []
-    for t in titles:
-        title = {key: t[i] for i, key in enumerate(keys)}
-        title["chapters"] = json.loads(title["chapters"])
-        title_dicts.append(title)
 
-    return title_dicts
+    for t in titles:
+        chapters = json.loads(t["chapters"])
+
+        # n+1 queries cuz I don't give a f- actually I do, but sqlite's cool with it:
+        # https://www.sqlite.org/np1queryprob.html
+        chapters_i_finished = run_sql_on_demand(
+            """
+            SELECT r.chapter_id
+            FROM read r
+                INNER JOIN chapter c ON c.id = r.chapter_id AND c.site = r.site
+            WHERE r.user_id = ?
+                AND c.title_id = ?
+                AND c.site = ?
+            ORDER BY c.num_major desc, c.num_minor desc;
+            """,
+            (user_id, t["id"], t["site"]),
+        )
+        # Cut off chapter list:
+        # only show chapters newer than the latest chapter that user has finished.
+        # Running a loop here instead of just picking the one latest finished chapter
+        # because source site may have deleted said chapter.
+        for finished_chapter_id in chapters_i_finished:
+            for i, ch in enumerate(chapters):
+                if finished_chapter_id == ch["id"]:
+                    chapters = chapters[:i]
+                    break
+
+        t["chapters"] = chapters
+
+    return sorted(titles, key=lambda t: len(t["chapters"]), reverse=True)
+
+
+def read(user_id, site, chapter_id):
+    run_sql(
+        """
+        INSERT INTO read (user_id, site, chapter_id) VALUES (?, ?, ?)
+        ON CONFLICT (user_id, site, chapter_id) DO UPDATE SET updated_at=datetime('now')
+        """,
+        (user_id, site, chapter_id),
+    )
+
+
+def unread(user_id, site, chapter_id):
+    run_sql(
+        "DELETE FROM read WHERE user_id=? AND site=? AND chapter_id=?;",
+        (user_id, site, chapter_id),
+    )
